from typing import TypeVar

actionKey = TypeVar("actionKey", str, int)
solverKey = TypeVar("solverKey", str, int)
agentKey = TypeVar("agentKey", str, int)


__all__: list[str] = ["actionKey", "solverKey", "agentKey"]



import logging
from typing import Callable, Dict, Iterable, Tuple, Type

from numpy import float64, int64, ndarray

from BanditAgents import BaseSolverHyperParameters, SimulationContext

from BanditAgents.src.agents.base_agent import BaseAgent
from BanditAgents.src.domain import actionKey, agentKey
from BanditAgents.src.domain.hyperparameters import SimulationParameters
from BanditAgents.src.solvers.solver import Solver


class SimulationAgent(BaseAgent):
    agent_id: agentKey
    context: SimulationContext
    solvers: list[Type[(Solver,)]]

    def __init__(
        self,
        actions: Iterable[
            Tuple[
                actionKey,
                Callable[[any], float] | Tuple[float, ...] | ndarray[float],
            ]
        ],
        solvers_hyperparameters: Iterable[Type[(BaseSolverHyperParameters)]],
        agent_id: agentKey = False,
    ) -> None:
        """_summary_

        Parameters
        ----------
        actions : Iterable[
            Tuple[
                actionKey,
                Callable[[any], float]
                |  Tuple[float, ...]
                |  ndarray[float], ]
            ]
            _description_
        solvers_hyperparameters : Iterable[Type[
            _description_
        agent_id : agentKey, optional
            _description_, by default False
        """
        self.logger: logging.Logger = logging.getLogger(__name__)

        super().__init__(agent_id=agent_id)

        self.context = SimulationContext(actions=actions)

        action_keys: Tuple[actionKey, ...] = tuple(
            action_key for action_key in self.context.action_dict.keys()
        )

        self.solvers = list(
            self._from_solver_hyperparameters_make_solver(
                action_keys=action_keys,
                solver_hyperparameters=solver_hyperparameters,
            )
            for solver_hyperparameters in solvers_hyperparameters
        )

    def run(
        self, simulation_parameters: SimulationParameters, as_dict: bool = True
    ) -> Iterable[
        Tuple[
            str,
            Dict[str, ndarray]
            | Tuple[
                ndarray[int64], ndarray[int64], ndarray[str], ndarray[float64]
            ],
        ]
    ]:
        """_summary_

        Parameters
        ----------
        simulation_parameters : SimulationParameters
            _description_
        as_dict : bool, optional
            _description_, by default True

        Returns
        -------
        Iterable[
            Tuple[
                str,
                Dict[str, any]
                | Tuple[ndarray[int64], ndarray[int64], ndarray[str], ndarray[float64]],
            ]
        ]
            _description_
        """
        simulation_run: Iterable[
            Tuple[
                str,
                Dict[str, any]
                | Tuple[
                    ndarray[int64],
                    ndarray[int64],
                    ndarray[str],
                    ndarray[float64],
                ],
            ]
        ] = self.context.run(
            solvers=self.solvers,
            as_dict=as_dict,
            **simulation_parameters.__dict__
        )

        return simulation_run



from typing import Dict, Iterable, List, Tuple
from numpy import ndarray
from pandas import DataFrame, concat
from Exemples.performance_exemples.basic_performance_comparison_exemple import (
    basic_performance_comparison_exemple,
)


def run_performance_exemples() -> DataFrame:
    simulation_results: Iterable[Tuple[str, Dict[str, ndarray]]] = (
        basic_performance_comparison_exemple()
    )

    sim_dfs: List[DataFrame] = []
    for solver_id, solver_simulation_result in simulation_results:
        solver_simulation_df = DataFrame(solver_simulation_result)
        solver_simulation_df["simulation"] = solver_id

        sim_dfs = [*sim_dfs, solver_simulation_df]

    simulation_df: DataFrame = concat(sim_dfs)

    return simulation_df


__all__: List[str] = [
    "run_performance_exemples",
    "basic_performance_comparison_exemple",
]



from typing import Iterable
from BanditAgents.src.contexts.deployable.deployable_context import (
    DeployableContext,
)
from BanditAgents.src.domain import actionKey


class EmptyContext(DeployableContext):
    def __init__(self, action_keys: Iterable[actionKey]) -> None:
        """_summary_

        Parameters
        ----------
        action_keys : Iterable[actionKey]
            keys associated to the actions that will be called
        """
        super().__init__(action_keys=action_keys)



from typing import List
import unittest
from unittest.mock import patch

from BanditAgents.src.solvers.epsilon_solver import EpsilonSolver
from BanditAgents.src.domain import actionKey
from BanditAgents.src.solvers.weight_solver import WeightSolver

WEIGHT_SOLVER_PATH: str = f'{EpsilonSolver.__module__}.{WeightSolver.__name__}'
WEIGHT_SOLVER_PREDICT_PATH: str = (
    f'{WEIGHT_SOLVER_PATH}.{WeightSolver.predict.__name__}'
)
WEIGHT_SOLVER_RANDOM_ACTION_PATH: str = (
    f'{WEIGHT_SOLVER_PATH}.{WeightSolver._random_action.__name__}'
)


def mock_predict() -> int:
    return 1


def mock_random() -> int:
    return 0


class TestEpsilonSolver(unittest.TestCase):
    epsilon_solver: EpsilonSolver
    mock_action_keys: List[actionKey]
    mock_optimistic_value: float
    mock_step_size: float
    mock_epsilon: float

    @patch(WEIGHT_SOLVER_PATH)
    def setUp(self, weight_solver) -> None:
        self.mock_action_keys = ["action_a", "action_b"]
        self.mock_optimistic_value = 1.0
        self.mock_step_size = 1.0
        self.mock_epsilon = 0.0

        self.epsilon_solver = EpsilonSolver(
            action_keys=self.mock_action_keys,
            optimistic_value=self.mock_optimistic_value,
            step_size=self.mock_step_size,
            epsilon=self.mock_epsilon,
        )

    @patch(WEIGHT_SOLVER_RANDOM_ACTION_PATH, side_effect=mock_random)
    @patch(WEIGHT_SOLVER_PREDICT_PATH, side_effect=mock_predict)
    def test_predict_succeed(self, predict, _random_action) -> None:
        self.epsilon_solver.__class__.__bases__[0].predict = predict
        self.epsilon_solver.__class__.__bases__[0]._random_action = (
            _random_action
        )

        action_index: int = self.epsilon_solver.predict()

        predict.assert_called_once()
        self.assertEqual(action_index, 1)

    @patch(WEIGHT_SOLVER_RANDOM_ACTION_PATH, side_effect=mock_random)
    @patch(WEIGHT_SOLVER_PREDICT_PATH, side_effect=mock_predict)
    def test_predict_with_random_action_succeed(self, predict, _random_action):
        self.epsilon_solver.__class__.__bases__[0].predict = predict
        self.epsilon_solver.__class__.__bases__[0]._random_action = (
            _random_action
        )
        mock_epsilon = 1.0

        self.epsilon_solver.epsilon = mock_epsilon

        action_index: int = self.epsilon_solver.predict()

        _random_action.assert_called_once()

        self.assertEqual(action_index, 0)



class InvalidActionDataException(Exception):
    def __init__(self, action_data: any) -> None:
        message = f"""
        action data is invalid:
            {action_data}
        """

        super().__init__(message)



from BanditAgents.src.agents.agent import Agent
from BanditAgents.src.solvers import (
    Solvers,
    EpsilonSolver,
    WeightSolver,
    UCBSolver,
    SamplingSolver,
    Solver,
)
from BanditAgents.src.contexts.simulation_context import SimulationContext

from BanditAgents.src.domain import actionKey, solverKey

from BanditAgents.src.domain.hyperparameters import (
    BaseSolverHyperParameters,
    SamplingSolverHyperParameters,
    WeightSolverHyperParameters,
    UCBSolverHyperParameters,
    SyncContextHyperParameters,
    DeployableContextHyperParameters,
    EmptyContextHyperParameters,
    SimulationParameters,
)

__all__: list[str] = [
    "Agent",
    "Solver",
    "EpsilonSolver",
    "SamplingSolver",
    "UCBSolver",
    "WeightSolver",
    "Solvers",
    "SimulationContext",
    "actionKey",
    "BaseSolverHyperParameters",
    "SamplingSolverHyperParameters",
    "WeightSolverHyperParameters",
    "UCBSolverHyperParameters",
    "SyncContextHyperParameters",
    "solverKey",
    "DeployableContextHyperParameters",
    "EmptyContextHyperParameters",
    "SimulationParameters",
]



from enum import Enum


class WeightSolverArgs(Enum):
    SOLVER_ID: str = "solver_id"
    ACTION_KEYS: str = "action_keys"
    OPTIMISTIC_VALUE: str = "optimistic_value"
    STEP_SIZE: str = "step_size"


class EpsilonSolverArgs(Enum):
    SOLVER_ID: str = WeightSolverArgs.SOLVER_ID.value
    ACTION_KEYS: str = WeightSolverArgs.ACTION_KEYS.value
    OPTIMISTIC_VALUE: str = WeightSolverArgs.OPTIMISTIC_VALUE.value
    STEP_SIZE: str = WeightSolverArgs.STEP_SIZE.value
    EPSILON: str = "epsilon"


class UCBSolverArgs(Enum):
    SOLVER_ID: str = WeightSolverArgs.SOLVER_ID.value
    ACTION_KEYS: str = WeightSolverArgs.ACTION_KEYS.value
    OPTIMISTIC_VALUE: str = WeightSolverArgs.OPTIMISTIC_VALUE.value
    STEP_SIZE: str = WeightSolverArgs.STEP_SIZE.value
    CONFIDENCE: str = "confidence"


class SamplingSolverArgs(Enum):
    SOLVER_ID: str = WeightSolverArgs.SOLVER_ID.value
    ACTION_KEYS: str = WeightSolverArgs.ACTION_KEYS.value
    N_SAMPLING: str = "n_sampling"
    MAX_SAMPLE_SIZE: str = "max_sample_size"



from typing import Iterable
from BanditAgents.src.domain import actionKey
from BanditAgents.src.contexts.context import Context


class DeployableContext(Context):
    action_keys: tuple[actionKey, ...]

    def __init__(self, action_keys: Iterable[actionKey]) -> None:
        self.action_keys = tuple(action_key for action_key in action_keys)

    def get_action_keys(self) -> tuple[actionKey, ...]:
        """Gets action keys

        Returns
        -------
        ndarray[actionKey]
            an array of action keys
        """
        return self.action_keys

    def info(self) -> dict[str, tuple[actionKey, ...]]:
        return dict(action_keys=self.action_keys)



from dataclasses import dataclass
from typing import Callable, Iterable, Tuple

from BanditAgents.src.domain import actionKey, solverKey


@dataclass
class BaseSolverHyperParameters:
    solver_id: solverKey = False


@dataclass
class SamplingSolverHyperParameters(BaseSolverHyperParameters):
    n_sampling: int = None
    max_sample_size: int = None


@dataclass
class WeightSolverHyperParameters(BaseSolverHyperParameters):
    optimistic_value: float = None
    step_size: float = None


@dataclass
class EpsilonSolverHyperParameters(WeightSolverHyperParameters):
    epsilon: float = None


@dataclass
class UCBSolverHyperParameters(WeightSolverHyperParameters):
    confidence: float = None


@dataclass
class DeployableContextHyperParameters:
    pass


@dataclass
class SyncContextHyperParameters(DeployableContextHyperParameters):
    actions: Iterable[Tuple[actionKey, Callable[[any], float]]]


@dataclass
class EmptyContextHyperParameters(DeployableContextHyperParameters):
    action_keys: Iterable[actionKey]


@dataclass
class SimulationParameters:
    n_steps: int
    steps_by_ticks: int


__all__: list[str] = [
    "BaseSolverHyperParameters",
    "SamplingSolverHyperParameters",
    "WeightSolverHyperParameters",
    "EpsilonSolverHyperParameters",
    "UCBSolverHyperParameters",
    "SyncContextHyperParameters",
    "SimulationParameters",
    "DeployableContextHyperParameters",
    "SyncContextHyperParameters",
    "EmptyContextHyperParameters",
]



import logging
from typing import Iterable

from numpy import empty, float64, int64, ndarray
from BanditAgents import Agent, actionKey
from scipy.stats import gamma

from BanditAgents.src.domain.hyperparameters import EmptyContextHyperParameters


def empty_context_agent_exemple(n_steps: int = 100):
    # Say we cannot execute the agent's choices in a reasonable delay.
    # The action execution will take more time than is allowed for the
    # timeout of your application and so the agent cannot wait all that time.
    # You will have to execute the action independently of the agent,
    # to do so you can make use of an EmptyContext object
    exemple_logger: logging.Logger = logging.getLogger(__name__)

    def action_a() -> float:
        return gamma.rvs(a=6.8, scale=0.1, loc=0, size=1)[0]

    def action_b() -> float:
        return gamma.rvs(a=2.2, scale=0.2, loc=0, size=1)[0]

    print(action_a.__name__)

    def execute_actions(action_keys: Iterable[actionKey]) -> ndarray[float64]:
        targets: ndarray[float64] = empty(len(action_keys))

        for i, action_key in enumerate(action_keys):
            if action_key == action_a.__name__:
                target: float = action_a()

            elif action_key == action_b.__name__:
                target = action_b()

            targets[i] = target

        return targets

    exemple_logger.debug(
        "Starting basic agent exemple\n"
        "---------------------------------------------------\n"
        "Initiating agent"
    )

    context_hyperparameters = EmptyContextHyperParameters(
        action_keys=('action_a', 'action_b')
    )
    agent: Agent = Agent(context_hyperparameters)

    indexes: ndarray[int64]
    targets: ndarray[float64]

    exemple_logger.debug(f"Agent initiated \n{agent.info()}")

    for i in range(n_steps):
        exemple_logger.debug(f"running step {i}")

        indexes = agent.act()
        action_keys: tuple[actionKey, ...] = (
            agent.solver.indexes_to_action_keys(indexes)
        )
        targets = execute_actions(action_keys)

        agent = agent.fit(x=indexes, y=targets)

        exemple_logger.debug(f"agent info is: {agent.info()}")

    exemple_logger.debug("---------------------------------------------------")



import logging
from typing import Callable, List, Tuple
from numpy import ndarray
from scipy.stats import gamma

from BanditAgents import Solvers, SamplingSolver, SimulationContext, solverKey


def sampling_solver_example() -> ndarray[float]:
    sampling_example_logger: logging.Logger = logging.getLogger(__name__)

    # First we need some actions (any function that returns a float)
    # For our situation here we will mock the actions by using some
    # gamma distributions
    action_a: Callable[[], float] = lambda: gamma.rvs(
        a=6.8, scale=0.1, loc=0, size=1
    )[0]
    action_b: Callable[[], float] = lambda: gamma.rvs(
        a=2.2, scale=0.2, loc=0, size=1
    )[0]

    action_keys: Tuple[str, str] = ("action_a", "action_b")
    sampling_example_logger.debug(action_keys)

    # We now instanciate the solver
    # as the name of the function suggest this solver is a ucb solver,
    # meaning it will use the upper bounds weighted function to
    # make its decisions
    sample_solver: SamplingSolver = Solvers().sampling_solver(
        action_keys=action_keys, n_sampling=1
    )
    sampling_example_logger.debug(sample_solver)

    # This list is used to identify the actions in the context
    actions: List[Tuple[str, Callable[[], float]]] = [
        ("action_a", action_a),
        ("action_b", action_b),
    ]

    sampling_example_logger.debug(actions)

    # We now instanciate the Context of the experiment
    # The context is an object that facilitates the use of the solvers
    # given a number of actions that can be called as functions.
    with SimulationContext(actions) as simulation:
        sampling_example_logger.debug(simulation)

        # the simulations will run for a 100 steps.
        # which means the solver will make a 100 choices
        # and action_a or b will be called a 100 times total
        targets: Tuple[solverKey, ndarray[float]] = next(
            simulation.run(
                n_steps=100,
                solvers=[sample_solver],
                steps_by_ticks=1,
                as_dict=True,
            )
        )

        sampling_example_logger.debug(targets)

    return targets[1]



from pandas import DataFrame, concat

from Exemples.basic_exemples.epsilon_solver_example import (
    epsilon_solver_example,
)
from Exemples.basic_exemples.sampling_solver_example import (
    sampling_solver_example,
)
from Exemples.basic_exemples.ucb_solver_example import ucb_solver_example
from Exemples.basic_exemples.weight_solver_example import weight_solver_example


def run_basic_exemples() -> DataFrame:
    ws_simulation_results = DataFrame(weight_solver_example())
    ws_simulation_results["simulation"] = "weighted"
    es_simulation_results = DataFrame(epsilon_solver_example())
    es_simulation_results["simulation"] = "epsilon"
    ucbs_simulation_result = DataFrame(ucb_solver_example())
    ucbs_simulation_result["simulation"] = "ucb"
    sampling_simulation_result = DataFrame(sampling_solver_example())
    sampling_simulation_result["simulation"] = "sampling"

    simulation_results: DataFrame = concat(
        [
            ws_simulation_results,
            es_simulation_results,
            ucbs_simulation_result,
            sampling_simulation_result,
        ]
    )

    return simulation_results


__all__ = ['run_basic_exemples']



from Exemples.agents_exemples.empty_context_agent_exemple import (
    empty_context_agent_exemple,
)
from Exemples.agents_exemples.basic_agent_exemple import basic_agent_exemple


def run_agents_exemples() -> None:
    basic_agent_exemple()
    empty_context_agent_exemple()


__all__ = ['run_agents_exemples']



from typing import Iterable
from numpy import random
from BanditAgents.src.solvers.weight_solver import WeightSolver
from BanditAgents.src.domain import actionKey, solverKey


class EpsilonSolver(WeightSolver):
    epsilon: float

    def __init__(
        self,
        action_keys: Iterable[actionKey],
        optimistic_value: float = 0.0,
        step_size: float = 1.0,
        epsilon: float = 1e-10,
        solver_id: solverKey = None,
    ) -> None:
        """_summary_

        Parameters
        ----------
        action_keys : Iterable[actionKey]
            _description_
        optimistic_value : float, optional
            _description_, by default 0.
        step_size : float, optional
            _description_, by default 1.
        epsilon : float, optional
            _description_, by default 1e-10
        solver_id: solverKey, optional
            _description_, by default None
        """
        super().__init__(
            action_keys=action_keys,
            optimistic_value=optimistic_value,
            step_size=step_size,
            solver_id=solver_id,
        )

        self.epsilon = epsilon

    def predict(self) -> int:
        """_summary_

        Returns
        -------
        int
            _description_
        """
        if random.random(1)[0] > self.epsilon:
            return super().predict()

        else:
            return self._random_action()



import logging
from typing import Any, Callable, Dict, Iterable, List, Self, Tuple, Type
from numpy import arange, array, dtype, empty, float64, int64, ndarray
from scipy.stats import gamma
from BanditAgents.src.contexts.context import Context
from BanditAgents.src.domain import actionKey
from BanditAgents.src.domain.context_action_args import MakeGammaActionArgs
from BanditAgents.src.exceptions.context_exceptions.invalid_action_data_exception import (
    InvalidActionDataException,
)
from BanditAgents.src.solvers.solver import Solver


class SimulationContextActionFactory(Context):
    def make_gamma_action_from_data(
        self, y: ndarray[float]
    ) -> Callable[[], float]:
        """Sometimes you might want to run a simulation
        with actions that try to mimic real data.

        This method facilitate this process

        Parameters
        ----------
        y : ndarray[float]
            Vector of values to compute the distribution

        Returns
        -------
        Callable[[], float]
            a function that calls a gamma distribution fitted on y
        """
        shape, loc, scale = gamma.fit(y)

        f: Callable[[], float] = self.make_gamma_action(
            alpha=shape, loc=loc, scale=scale
        )

        return f

    def make_gamma_action(
        self, alpha: float, loc: float = None, scale: float = None
    ) -> Callable[[], float]:
        """Generates a gamma.rvs function based of alpha, loc, scale

        Parameters
        ----------
        alpha : float
            shape parameter of the distribution
        loc : float, optional
            loc parameter of the distribution, by default None
        scale : float, optional
            scale parameter of the distribution, by default None

        Returns
        -------
        Callable[[], float]
            a gamma rvs that will always be executed with parameters
                alpha, loc and scale
        """
        g_kwargs = dict(a=alpha)

        if loc is not None:
            g_kwargs[MakeGammaActionArgs.LOC.value] = loc

        if scale is not None:
            g_kwargs[MakeGammaActionArgs.SCALE.value] = scale

        f: Callable[[], float] = lambda: gamma.rvs(**g_kwargs, size=1)[0]

        return f


class SimulationContext:
    action_dict: Dict[actionKey, Callable[[any], float]]
    simulation_context_action_factory: SimulationContextActionFactory

    def __init__(
        self,
        actions: Iterable[
            Tuple[
                actionKey,
                Callable[[any], float] | Tuple[float, ...] | ndarray[float],
            ]
        ],
    ) -> None:
        """Constructor of the simulation object

        Parameters
        ----------
        actions : Iterable[
            Tuple[actionKey, Callable[[any], float]
            |  Tuple[float, ...]
            |  ndarray[float]]
        ]
            _An Iterable, could be a list, a tuple, a generator, etc. of pairs.

            - The first member of every pair must be an actionKey.
            - The second member can be a Callable that must return a float,
                a tuple containing alpha, loc and scale parameters
                for a gamma distribution,
                or an array of floats that will be used
                to fit a gamma distribution on them
        """
        self.logger: logging.Logger = logging.getLogger(__name__)

        self.simulation_context_action_factory = (
            SimulationContextActionFactory()
        )
        self.action_dict = dict()

        self.add_actions(actions)

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.logger.debug(traceback)

        return

    def act(
        self, targets: ndarray[float64], step: int, action_index: int
    ) -> float:
        """Execute a gievn solver's decision

        Parameters
        ----------
        targets : ndarray[float64]
            Matrix n_steps by n_actions containing results of action j at step i
        step : int
            Row selector in the target matrix
        action_index : int
            Column selector in the target matrix

        Returns
        -------
        float
            value found at target[step, action_index]
        """
        target: float = targets[step, action_index]

        return target

    def add_action(
        self,
        action_key: actionKey,
        action_data: (
            Callable[[any], float] | Tuple[float, ...] | ndarray[float]
        ),
        is_return: bool = True,
    ) -> Self | None:
        """Add an action to the context

        Parameters
        ----------
        action_key : actionKey
            identifier of the action
        action_data : Callable[[any], float]
            | Tuple[float, ...]
            | ndarray[float]
            if action_data is callable, add the action to the context
            if action_data  is a tuple of 3 floats it will create a
                gamma function that uses the three floats as
                shape, loc, scale respectively
            if action_data  is an array of float it will return a gamma
                function fitted on the array
        is_return : bool, optional
            if is return, returns the Simulation context itself,
            otherwise returns None, by default True

        Returns
        -------
        Self | None
            if is_return is True returns self otherwise return None
        """
        if callable(action_data):
            self.logger.debug(f"action {action_key} is a custom function")

            action_func: Callable[[any], float] = action_data

        elif (type(action_data) is tuple) and (len(action_data) == 3):
            self.logger.debug(
                f"action {action_key} will bemade based on following params {action_data}"
            )

            action_func: Callable[[], float] = (
                self.simulation_context_action_factory.make_gamma_action(
                    *action_data
                )
            )

        elif isinstance(action_data, ndarray[float]):
            self.logger.debug(
                f"action {action_key} will be a gamma fitted on given data"
            )

            factory: SimulationContextActionFactory = (
                self.simulation_context_action_factory
            )
            action_func: Callable[[], float] = (
                factory.make_gamma_action_from_data(action_data)
            )

        else:
            raise InvalidActionDataException(action_data)

        self.logger.debug(action_func)

        self.action_dict[action_key] = action_func

        if is_return:
            return self

        return

    def add_actions(
        self,
        actions: Iterable[
            Tuple[
                actionKey,
                Callable[[any], float] | Tuple[float, ...] | ndarray[float],
            ]
        ],
    ) -> Self:
        """Through this method you can add actions to the Simulation

        Parameters
        ----------
        actions : Iterable[
            Tuple[
                actionKey,
                Callable[[any], float]
                |  Tuple[float, ...]
                |  ndarray[float]
            ]
        ]
            An Iterable, could be a list, a tuple, a generator, etc. of pairs.

            - The first member of every pair must be an actionKey.
            - The second member can be a Callable that must return a float,
                a tuple containing alpha, loc and scale parameters
                for a gamma distribution,
                or an array of floats that will be used
                to fit a gamma distribution on them

        Returns
        -------
        Self
            The SimulationContext with the added actions
        """
        for action_key, action_func in actions:
            self.add_action(action_key, action_func, False)

        return self

    def run(
        self,
        n_steps: int,
        solvers: List[Type[(Solver,)]],
        steps_by_ticks: int = 1,
        as_dict: bool = False,
    ) -> Iterable[
        Tuple[
            str,
            Dict[str, any]
            | Tuple[
                ndarray[int64], ndarray[int64], ndarray[str], ndarray[float64]
            ],
        ]
    ]:
        """Runs the silumation on given solvers

        Parameters
        ----------
        n_steps : int
            Number of steps that the simulation will execute
        solvers : List[Type[(BaseSolver,)]]
            List of solvers that will interact with the simulation
        steps_by_ticks : int, optional
            Number of steps between fits, by default 1
        as_dict : bool, optional
            define whether or not you want to yielded objects
            to be arrays or dictionnaries, by default False

        Returns
        -------
        Tuple[ndarray[int64], ndarray[int64], ndarray[str], ndarray[float64]]
        | Dict[str, ndarray]
            results of each solver's simulation
        """
        action_keys: Tuple[actionKey] = solvers[0].action_keys

        # insure that action keys are aligned
        assert all(
            [solver.action_keys == action_keys for solver in solvers[1:]]
        )

        steps: ndarray[int64] = arange(0, n_steps)
        targets: ndarray[float64] = empty((n_steps, len(self.action_dict)))

        for step in range(n_steps):
            targets[step] = array([self.action_dict[k]() for k in action_keys])

        for solver in solvers:
            yield self._run_solver_simulation(
                n_steps=n_steps,
                steps_by_ticks=steps_by_ticks,
                steps=steps,
                targets=targets,
                as_dict=as_dict,
                solver=solver,
            )

    def _run_solver_simulation(
        self,
        n_steps: int,
        steps_by_ticks: int,
        steps: ndarray[int64],
        targets: ndarray[float64],
        as_dict: bool,
        solver: Type[(Solver,)],
    ) -> Tuple[
        str,
        Dict[str, any]
        | Tuple[
            ndarray[int64], ndarray[int64], ndarray[str], ndarray[float64]
        ],
    ]:
        """Run simulation for one solver

        Parameters
        ----------
        n_steps : int
            Number of steps that the simulation will execute
        steps_by_ticks : int
            Number of steps between fits
        steps : ndarray[int64]
            Array of steps
        targets : ndarray[float64]
            Matrix n_steps by n_actions containing results of action j at step i
        as_dict : bool
            define whether or not you want to yielded objects
            to be arrays or dictionnaries
        solver : Type[
            Solver that will interact with the simulation

        Returns
        -------
        Tuple[
            str,
            Dict[str, any]
            | Tuple[ndarray[int64], ndarray[int64], ndarray[str], ndarray[float64]],
        ]
            Results of the solver's simulation
        """
        self.logger.debug(f"Running simulation for {solver.solver_id}")

        last_training_index: int = 0
        solver_targets: ndarray[float64] = empty(n_steps)
        action_keys: ndarray[str] = empty(n_steps, dtype="<U100")
        indexes: ndarray[int64] = empty(n_steps, dtype=int64)

        for step in range(n_steps):
            self._take_a_step(
                step=step,
                steps=steps,
                indexes=indexes,
                steps_by_ticks=steps_by_ticks,
                n_steps=n_steps,
                action_keys=action_keys,
                targets=targets,
                solver_targets=solver_targets,
                solver=solver,
                last_training_index=last_training_index,
            )

        self.logger.debug(
            "----------------------------------------------------------"
            "---------------------------------\nRun completed!\n"
        )

        if as_dict:
            results: Dict[str, ndarray] = {
                "steps": steps,
                "action_indexes": indexes,
                "action_keys": action_keys,
                "targets": solver_targets,
            }

        else:
            results = (steps, indexes, action_keys, targets)

        return solver.solver_id, results

    def _take_a_step(
        self,
        step: int,
        steps: ndarray[int64],
        indexes: ndarray[int64],
        steps_by_ticks: int,
        n_steps: int,
        action_keys: ndarray[str],
        targets: ndarray[float64],
        solver_targets: ndarray[float64],
        solver: Type[(Solver,)],
        last_training_index: int,
    ) -> None:
        """Run one step of the simulation

        Parameters
        ----------
        step : int
            Current step of the simulation
        steps : ndarray[int64]
            Array of steps
        indexes : ndarray[int64]
            Array of solver's decisions
        steps_by_ticks : int
            Number of steps between fits
        n_steps : int
            Number of steps that the simulation will execute
        action_keys : ndarray[str]
            array of action identifiers
        targets : ndarray[float64]
            Matrix n_steps by n_actions containing results of action j at step i
        solver_targets : ndarray[float64]
            results of the solver's decisions
        solver : Type[(BaseSolver,)]
            Solver interacting with the simulation
        last_training_index : int
            Last step at which the solver was trained
        """
        self.logger.debug(
            f"Simulation step {step}\n---------------------------------------"
            "----------------------------------------------------"
        )
        action_index: int = solver.predict()
        indexes[step] = action_index

        if (step % steps_by_ticks == 0 or step == n_steps - 1) and step != 0:
            self._fit_step(
                step=step,
                steps=steps,
                indexes=indexes,
                steps_by_ticks=steps_by_ticks,
                n_steps=n_steps,
                action_keys=action_keys,
                targets=targets,
                solver_targets=solver_targets,
                solver=solver,
                last_training_index=last_training_index,
            )

    def _fit_step(
        self,
        step: int,
        steps: ndarray[int64],
        indexes: ndarray[int64],
        steps_by_ticks: int,
        n_steps: int,
        action_keys: ndarray[str],
        targets: ndarray[float64],
        solver_targets: ndarray[float64],
        solver: Type[(Solver,)],
        last_training_index: int,
    ) -> None:
        """Execute fitting part of a step

        Parameters
        ----------
        step : int
            Current step of the simulation
        steps : ndarray[int64]
            Array of steps
        indexes : ndarray[int64]
            Array of solver's decisions
        steps_by_ticks : int
            Number of steps between fits
        n_steps : int
            Number of steps that the simulation will execute
        action_keys : ndarray[str]
            array of action identifiers
        targets : ndarray[float64]
            Matrix n_steps by n_actions containing results of action j at step i
        solver_targets : ndarray[float64]
            results of the solver's decisions
        solver : Type[(BaseSolver,)]
            Solver interacting with the simulation
        last_training_index : int
            Last step at which the solver was trained
        """
        self.logger.debug(f"step {step} is a training step")

        difference: int = (
            (step - last_training_index)
            if step == n_steps - 1
            else steps_by_ticks
        )
        start_index: int = step - difference
        end_index: int = step + 1 if step == n_steps - 1 else step

        self.logger.debug(
            "training on targets indexes: " f"{start_index} to {end_index}"
        )

        steps_to_execute: ndarray[int64] = steps[start_index:end_index]
        indexes_to_execute: ndarray[int64] = indexes[start_index:end_index]
        self.logger.debug(
            f"{solver.solver_id}'s decision indexes were {indexes_to_execute}"
        )

        action_keys_to_execute: Tuple[actionKey] = array(
            [
                action_key
                for action_key in solver.indexes_to_action_keys(
                    indexes_to_execute
                )
            ],
            dtype="<U100",
        )
        self.logger.debug(
            f"Which corresponds to actions {action_keys_to_execute}"
        )

        action_keys[start_index:end_index] = action_keys_to_execute

        self.logger.debug("Executing decisions")

        tick_targets: ndarray[Any, dtype[Any]] = array(
            [
                self.act(targets, step_to_execute, action_index)
                for step_to_execute, action_index in zip(
                    steps_to_execute, indexes_to_execute
                )
            ]
        )
        self.logger.debug(
            f"Decisions wielded following targets {tick_targets}"
        )

        solver_targets[start_index:end_index] = tick_targets

        self.logger.debug(f"Fitting solver {solver.solver_id} with targets")
        solver = solver.fit(x=indexes_to_execute, y=tick_targets)
        self.logger.debug(
            "the training resulted in the following weights "
            f"{solver.weights}"
        )

        last_training_index = step



from typing import Callable, Self, Tuple

from numpy import ndarray

from BanditAgents.src.contexts.simulation_context import SimulationContext
from BanditAgents.src.domain import actionKey


class SimulationContextBuilder:
    actions: Tuple[
        Tuple[actionKey, Callable[[]], float]
        | Tuple[float, float, float]
        | ndarray[float],
        ...,
    ]

    def __init__(self) -> None:
        self.actions = None

    def with_custom_actions(
        self,
        actions: Tuple[
            Tuple[actionKey, Callable[[]], float]
            | Tuple[float, float, float]
            | ndarray[float],
            ...,
        ],
    ) -> Self:
        self.actions = actions

        return self

    def with_default_actions(self) -> Self:
        actions: tuple[
            tuple[str, Callable[[], float]], tuple[str, Callable[[], float]]
        ] = (("action_a", lambda: 4.0), ("action_b", lambda: 2.0))

        self.actions = actions

        return self

    def build(self) -> SimulationContext:
        if self.actions is None:
            self.with_default_actions()

        simulation_context = SimulationContext(actions=self.actions)

        return simulation_context



from typing import Dict, Iterable
from BanditAgents.src.domain.solver_args import (
    EpsilonSolverArgs,
    SamplingSolverArgs,
    UCBSolverArgs,
    WeightSolverArgs,
)
from BanditAgents.src.solvers.solver import Solver
from BanditAgents.src.solvers.epsilon_solver import EpsilonSolver
from BanditAgents.src.solvers.sampling_solver import SamplingSolver
from BanditAgents.src.solvers.ucb_solver import UCBSolver
from BanditAgents.src.solvers.weight_solver import WeightSolver
from BanditAgents.src.domain import actionKey, solverKey


class Solvers:
    def epsilon_solver(
        self,
        action_keys: Iterable[actionKey],
        solver_id: solverKey = None,
        optimistic_value: float = None,
        step_size: float = None,
        epsilon: float = None,
    ) -> EpsilonSolver:
        es_kwargs: Dict[str, Iterable[actionKey] | float] = (
            self._make_epsilon_solver_kwargs(
                action_keys=action_keys,
                solver_id=solver_id,
                optimistic_value=optimistic_value,
                step_size=step_size,
                epsilon=epsilon,
            )
        )

        epsilon_solver = EpsilonSolver(**es_kwargs)

        return epsilon_solver

    def sampling_solver(
        self,
        action_keys: Iterable[actionKey],
        solver_id: solverKey = None,
        n_sampling: int = None,
        max_sample_size: int = None,
    ) -> SamplingSolver:
        sampling_kwargs: Dict[str, Iterable[actionKey] | float | int] = (
            self._make_sampling_solver_kwargs(
                action_keys=action_keys,
                solver_id=solver_id,
                n_sampling=n_sampling,
                max_sample_size=max_sample_size,
            )
        )

        sampling_solver = SamplingSolver(**sampling_kwargs)

        return sampling_solver

    def ucb_solver(
        self,
        action_keys: Iterable[actionKey],
        solver_id: solverKey = None,
        optimistic_value: float = None,
        step_size: float = None,
        confidence: float = None,
    ) -> UCBSolver:
        ucb_kwargs: Dict[str, Iterable[actionKey] | float] = (
            self._make_ucb_solver_kwargs(
                action_keys=action_keys,
                solver_id=solver_id,
                optimistic_value=optimistic_value,
                step_size=step_size,
                confidence=confidence,
            )
        )

        ucb_solver = UCBSolver(**ucb_kwargs)

        return ucb_solver

    def weight_solver(
        self,
        action_keys: Iterable[actionKey],
        solver_id: solverKey = None,
        optimistic_value: float = None,
        step_size: float = None,
    ) -> WeightSolver:
        ws_kwargs: Dict[str, Iterable[actionKey] | float] = (
            self._make_weight_solver_kwargs(
                action_keys=action_keys,
                solver_id=solver_id,
                optimistic_value=optimistic_value,
                step_size=step_size,
            )
        )

        weight_solver = WeightSolver(**ws_kwargs)

        return weight_solver

    def _make_epsilon_solver_kwargs(
        self,
        action_keys: Iterable[actionKey],
        solver_id: solverKey = None,
        optimistic_value: float = None,
        step_size: float = None,
        epsilon: float = None,
    ) -> Dict[str, Iterable[actionKey] | float]:
        es_kwargs: Dict[str, Iterable[actionKey] | float] = (
            self._make_weight_solver_kwargs(
                action_keys=action_keys,
                solver_id=solver_id,
                optimistic_value=optimistic_value,
                step_size=step_size,
            )
        )

        if epsilon is not None:
            es_kwargs[EpsilonSolverArgs.EPSILON.value] = epsilon

        return es_kwargs

    def _make_sampling_solver_kwargs(
        self,
        action_keys: Iterable[actionKey],
        solver_id: solverKey = None,
        n_sampling: int = None,
        max_sample_size: int = None,
    ) -> Dict[str, Iterable[actionKey] | float]:
        sampling_kwargs = dict(action_keys=action_keys)

        if solver_id is not None:
            sampling_kwargs[WeightSolverArgs.SOLVER_ID.value] = solver_id

        if n_sampling is not None:
            sampling_kwargs[SamplingSolverArgs.N_SAMPLING.value] = n_sampling

        if max_sample_size is not None:
            sampling_kwargs[SamplingSolverArgs.MAX_SAMPLE_SIZE.value] = (
                max_sample_size
            )

        return sampling_kwargs

    def _make_ucb_solver_kwargs(
        self,
        action_keys: Iterable[actionKey],
        solver_id: solverKey = None,
        optimistic_value: float = None,
        step_size: float = None,
        confidence: float = None,
    ) -> Dict[str, Iterable[actionKey] | float]:
        ucb_kwargs: Dict[str, Iterable[actionKey] | float] = (
            self._make_weight_solver_kwargs(
                action_keys=action_keys,
                solver_id=solver_id,
                optimistic_value=optimistic_value,
                step_size=step_size,
            )
        )

        if confidence is not None:
            ucb_kwargs[UCBSolverArgs.CONFIDENCE.value] = confidence

        return ucb_kwargs

    def _make_weight_solver_kwargs(
        self,
        action_keys: Iterable[actionKey],
        solver_id: solverKey = None,
        optimistic_value: float = None,
        step_size: float = None,
    ) -> Dict[str, Iterable[actionKey] | float]:
        ws_kwargs = dict(action_keys=action_keys)

        if solver_id is not None:
            ws_kwargs[WeightSolverArgs.SOLVER_ID.value] = solver_id

        if optimistic_value is not None:
            ws_kwargs[WeightSolverArgs.OPTIMISTIC_VALUE.value] = (
                optimistic_value
            )

        if step_size is not None:
            ws_kwargs[WeightSolverArgs.STEP_SIZE.value] = step_size

        return ws_kwargs


__all__: list[str] = [
    "Solver",
    "EpsilonSolver",
    "SamplingSolver",
    "UCBSolver",
    "WeightSolver",
    "Solvers",
]



import logging
from typing import Any, Dict, Generator, Iterable, Self

from numpy import array, ndarray, random
from BanditAgents.src.solvers.solver import Solver
from BanditAgents.src.domain import actionKey, solverKey


class WeightSolver(Solver):
    optimistic_value: float
    step_size: float
    weights: ndarray[float]
    is_trained: bool

    def __init__(
        self,
        action_keys: Iterable[actionKey],
        optimistic_value: float = 0.0,
        step_size: float = 1.0,
        solver_id: solverKey = None,
    ) -> None:
        """_summary_

        Parameters
        ----------
        action_keys : Iterable[actionKey]
            _description_
        optimistic_value : float, optional
            _description_, by default 0.
        step_size : float, optional
            _description_, by default 1.
        """
        self.logger: logging.Logger = logging.getLogger(__name__)

        super().__init__(action_keys=action_keys, solver_id=solver_id)

        self.step_size = step_size
        self.optimistic_value = optimistic_value
        self.is_trained = False

        self._init_weights()

    def action_keys_to_indexes(
        self, action_keys: Iterable[actionKey]
    ) -> ndarray[int]:
        """_summary_

        Parameters
        ----------
        action_keys : Iterable[actionKey]
            _description_

        Returns
        -------
        ndarray[int]
            _description_
        """
        action_keys_index_dict: Dict[actionKey, int] = {
            action_key: i for i, action_key in enumerate(self.action_keys)
        }

        action_keys_indexes: ndarray[int] = array(
            [
                int(action_keys_index_dict[action_key])
                for action_key in action_keys
            ]
        )

        return action_keys_indexes

    def fit(self, x: ndarray[int], y: ndarray[float]) -> Self:
        """_summary_

        Parameters
        ----------
        x : ndarray[int]
            _description_
        y : ndarray[float]
            _description_

        Returns
        -------
        Self
            _description_
        """
        training_complete = tuple(self._steps(x, y))

        if all(training_complete):
            if not self.is_trained:
                self.is_trained = True

            return self

        else:
            return self

    def predict(self) -> int:
        """_summary_

        Returns
        -------
        int
            _description_
        """
        if self.is_trained:
            return self.weights.argmax()

        else:
            return self._random_action()

    def _init_weights(self) -> None:
        """_summary_"""
        self.weights = array([self.optimistic_value for _ in self.action_keys])

    def _steps(
        self, x: ndarray[int], y: ndarray[float]
    ) -> Generator[bool, Any, None]:
        """_summary_

        Parameters
        ----------
        x : _type_
            _description_
        y : _type_
            _description_

        Yields
        ------
        Generator[bool, Any, None]
            _description_
        """
        for action_index, target in zip(x, y):
            yield self._step(target=target, action_index=action_index)

    def _step(self, target: float, action_index: int) -> bool:
        """_summary_

        Parameters
        ----------
        target : float
            _description_
        action_index : int
            _description_

        Returns
        -------
        bool
            _description_
        """
        self.weights[action_index] = self._compute_weight(
            weight=self.weights[action_index],
            target=target,
            step_size_value=self.step_size,
        )

        return True

    def _random_action(self) -> int:
        """_summary_

        Returns
        -------
        int
            _description_
        """
        return random.randint(self.weights.size, size=1)[0]

    @staticmethod
    def _compute_weight(
        weight: float, target: float, step_size_value: float
    ) -> float:
        """_summary_

        Parameters
        ----------
        weight : float
            _description_
        target : float
            _description_
        step_size_value : float
            _description_

        Returns
        -------
        float
            _description_
        """
        new_weight: float = weight + step_size_value * (target - weight)

        return new_weight



import logging
from typing import Callable, Dict, Tuple, Type
from uuid import uuid4

from BanditAgents.src.solvers import Solvers

from BanditAgents.src.domain import actionKey, agentKey
from BanditAgents.src.domain.hyperparameters import (
    BaseSolverHyperParameters,
    SamplingSolverHyperParameters,
    UCBSolverHyperParameters,
    WeightSolverHyperParameters,
)

from BanditAgents.src.domain.hyperparameters import (
    EpsilonSolverHyperParameters,
)
from BanditAgents.src.solvers.solver import Solver


class BaseAgent:
    agent_id: agentKey
    solvers: Solvers = Solvers()
    solvers_dict: Dict[str, Callable[[any], Type[(Solver,)]]] = {
        EpsilonSolverHyperParameters.__name__: solvers.epsilon_solver,
        SamplingSolverHyperParameters.__name__: solvers.sampling_solver,
        UCBSolverHyperParameters.__name__: solvers.ucb_solver,
        WeightSolverHyperParameters.__name__: solvers.weight_solver,
    }

    def __init__(self, agent_id: agentKey = False) -> None:
        self.logger: logging.Logger = logging.getLogger(__name__)

        self.agent_id = agent_id if agent_id else uuid4()

    def _from_solver_hyperparameters_make_solver(
        self,
        action_keys: Tuple[actionKey],
        solver_hyperparameters: Type[(BaseSolverHyperParameters,)],
    ) -> Type[(Solver)]:
        self.logger.debug(
            f"making solver from hyperparameters {solver_hyperparameters}"
        )

        solver: Type[(Solver,)] = self.solvers_dict[
            type(solver_hyperparameters).__name__
        ](action_keys=action_keys, **solver_hyperparameters.__dict__)

        return solver



from typing import Dict, Iterable, List, Tuple, Type

from numpy import ndarray
from BanditAgents import (
    BaseSolverHyperParameters,
    SamplingSolverHyperParameters,
    UCBSolverHyperParameters,
    WeightSolverHyperParameters,
)
from BanditAgents.src.agents.simulation_agent import SimulationAgent
from BanditAgents.src.domain.hyperparameters import (
    EpsilonSolverHyperParameters,
    SimulationParameters,
)


def basic_performance_comparison_exemple() -> (
    Iterable[Tuple[str, Dict[str, ndarray]]]
):
    actions: Tuple[Tuple[str, Tuple[float, float, float]]] = (
        ("action_a", (6.8, 0, 0.1)),
        ("action_b", (2.2, 0, 0.2)),
    )

    epsilon_solvere_hyperparams = EpsilonSolverHyperParameters(
        solver_id="epsilon_1"
    )
    sampling_solver_hyperparams = SamplingSolverHyperParameters(
        solver_id="sampling_1"
    )
    ucb_solver_hyperparams = UCBSolverHyperParameters(solver_id="ucb_1")
    weight_solver_hyperparams = WeightSolverHyperParameters(
        solver_id="weight_1"
    )

    solvers_hyperparams: List[Type[(BaseSolverHyperParameters,)]] = [
        epsilon_solvere_hyperparams,
        sampling_solver_hyperparams,
        ucb_solver_hyperparams,
        weight_solver_hyperparams,
    ]

    agent = SimulationAgent(
        actions=actions, solvers_hyperparameters=solvers_hyperparams
    )

    simulation_params = SimulationParameters(n_steps=1000, steps_by_ticks=5)

    simulation_results: Iterable[Tuple[str, Dict[str, ndarray]]] = agent.run(
        simulation_parameters=simulation_params
    )

    return simulation_results



from typing import Dict, Generator, Iterable, Self, Tuple

from numpy import float64, ndarray, zeros
from uuid import uuid4

from BanditAgents.src.domain import actionKey, solverKey


class Solver:
    solver_id: solverKey
    action_keys: Tuple[actionKey]
    weights: ndarray[float64]

    def __init__(
        self,
        action_keys: Iterable[actionKey],
        solver_id: solverKey = None,
        *args,
        **kwargs
    ) -> None:
        """Constructor of the Solver class

        This class is a virtual class that should b inherited by all solver classes

        Parameters
        ----------
        action_keys : Iterable[actionKey]
            keys associated to the actions the solver will operate
        """
        if solver_id is not None:
            self.solver_id = solver_id

        else:
            self.solver_id = uuid4()

        self.action_keys = tuple(ac for ac in action_keys)
        self.weights = zeros(len(self.action_keys))

    def fit(self, *args, **kwargs) -> Self:
        """Signature of the fit method, all Solvers must have a fit method implemented

        This method should be what updates the solver's weights
        given information based on its previous actions

        Returns
        -------
        Self
            returns self after training
        """

    def indexes_to_action_keys(
        self, indexes: Iterable[int]
    ) -> Tuple[actionKey, ...]:
        """convert an iterable of action indexes into a tuple of action keys

        Parameters
        ----------
        indexes : Iterable[int]
            Iterable of action indexes

        Returns
        -------
        Tuple[actionKey, ...]
            Tuple of action keys associated to every index found in indexes
        """
        action_keys = tuple(self.action_keys[index] for index in indexes)

        return action_keys

    def info(self) -> Dict[str, any]:
        """Allows access to general information about the solver

        Returns
        -------
        Dict[str, any]
            returns a dictionnary containing the information
        """
        solver_info = dict(action_keys=self.action_keys, weights=self.weights)

        return solver_info

    def predict(self) -> int:
        """Signature of the predict method, all Solvers must have a predict method implemented

        This method produces the choice of action when the solver is called

        Returns
        -------
        int
            action index of the chosen action
        """

    def _step(self, *args, **kwargs) -> Generator[bool, any, None]:
        """Signature of the _step method, all Solvers must have a _step method implemented

        This method should be executed in fit tu update the weights

        Yields
        ------
        Generator[bool, any, None]
            boolean that represents if step was completed successfully
        """



import logging
from typing import Callable, Dict, Self, Tuple, Type

from numpy import empty, float64, int64, ndarray

from BanditAgents.src.agents.base_agent import BaseAgent
from BanditAgents.src.contexts.deployable.deployable_context import (
    DeployableContext,
)
from BanditAgents.src.contexts.deployable.empty_context import EmptyContext
from BanditAgents.src.domain import agentKey
from BanditAgents.src.domain.hyperparameters import (
    BaseSolverHyperParameters,
    DeployableContextHyperParameters,
    EmptyContextHyperParameters,
    SyncContextHyperParameters,
    SamplingSolverHyperParameters,
)
from BanditAgents.src.solvers import Solver
from BanditAgents.src.contexts.deployable.sync_context import SyncContext


class Agent(BaseAgent):
    actions_between_fits: int
    context: Type[(DeployableContext,)]
    solver: Type[(Solver,)]
    contexts_dict: Dict[str, Callable[[any], Type[(SyncContext,)]]] = {
        SyncContextHyperParameters.__name__: SyncContext,
        EmptyContextHyperParameters.__name__: EmptyContext,
    }

    def __init__(
        self,
        context_hyperparameters: Type[(DeployableContextHyperParameters,)],
        actions_between_fits: int = 1,
        solver_hyperparameters: Type[
            (BaseSolverHyperParameters,)
        ] = SamplingSolverHyperParameters(),
        agent_id: agentKey = False,
    ) -> None:
        """Constructor to instanciate the agent

        Parameters
        ----------
        context_hyperparameters : Type[(DeployableContextHyperParameters,)]
            Hyperparameters of the context
        actions_between_fits : int, optional
            Number of actions to execute before fitting the solver, by default 1
        solver_hyperparameters : Type[(BaseSolverHyperParameters,)], optional
            Hyperparameters of the solver,
            can be EpsilonSolverHyperParameters,
            SamplingSolverHyperParameters,
            UCBSolverHyperParameters or
            WeightSolverHyperParameters
            by default SamplingSolverHyperParameters()
        agent_id : agentKey, optional
            Id of the agent if false it will be a UUID, by default False
        """
        self.logger = logging.getLogger(__name__)

        super().__init__(agent_id=agent_id)

        self.actions_between_fits = actions_between_fits
        self.context = self._from_context_hyperparameters_make_context(
            context_hyperparameters=context_hyperparameters
        )
        self.solver = self._from_solver_hyperparameters_make_solver(
            action_keys=self.context.get_action_keys(),
            solver_hyperparameters=solver_hyperparameters,
        )

    def act(
        self, *args, **kwargs
    ) -> Tuple[ndarray[int64], ndarray[float64]] | ndarray[int64]:
        """Let the solver take action on the context

        Returns
        -------
        Tuple[ndarray[int64], ndarray[float64]]
            list of action ids and associated targets
        """
        is_empty_context: bool = isinstance(self.context, EmptyContext)

        results: Tuple[ndarray[int64], ndarray[float64]] | ndarray[int64]

        if is_empty_context:
            results = self.predict(*args, **kwargs)

        else:
            action_indexes: ndarray[int64] = empty(
                self.actions_between_fits, dtype=int64
            )

            targets: ndarray[float64] = empty(self.actions_between_fits)

            for i in range(self.actions_between_fits):
                action_index: int = self.solver.predict()
                action_indexes[i] = action_index
                targets[i] = self.context.execute(
                    action_index=action_index, *args, **kwargs
                )

            results = (action_indexes, targets)

        return results

    def predict(self, *args, **kwargs) -> ndarray[int64]:
        action_indexes: ndarray[int64] = empty(
            self.actions_between_fits, dtype=int64
        )

        for i in range(self.actions_between_fits):
            action_index: int = self.solver.predict()
            action_indexes[i] = action_index

        return action_indexes

    def fit(self, *args, **kwargs) -> Self:
        """Fit the solver

        Returns
        -------
        Self
            returns the newly fitted agent
        """
        self.solver.fit(*args, **kwargs)

        return self

    def info(self) -> Dict[str, any]:
        """Produces information about the agent

        Returns
        -------
        Dict[str, any]
            Information of the agent as a dictionary
        """
        agent_info = dict(
            context_info=self.context.info(), solver_info=self.solver.info()
        )

        return agent_info

    def _from_context_hyperparameters_make_context(
        self,
        context_hyperparameters: Type[(DeployableContextHyperParameters,)],
    ) -> Type[(DeployableContextHyperParameters,)]:
        """Makes a context from a context hyperparameter object

        Returns
        -------
        Type[(DeployableContextHyperParameters,)]
            An object that inherited from deployable context
        """
        context: Type[(DeployableContext,)] = self.contexts_dict[
            type(context_hyperparameters).__name__
        ](**context_hyperparameters.__dict__)

        return context



from typing import Dict, List, Tuple
import unittest

from numpy import array, float64, ndarray

from BanditAgents import Solver, actionKey


class TestBaseSolver(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_action_keys: List[actionKey] = ["action_a", "action_b"]
        self.base_solver: Solver = Solver(self.mock_action_keys)

    def test_indexes_to_action_keys_succeed(self) -> None:
        expected_action_keys: Tuple[actionKey, ...] = tuple(
            self.mock_action_keys
        )
        action_keys: Tuple[actionKey, ...] = (
            self.base_solver.indexes_to_action_keys([0, 1])
        )

        self.assertEqual(action_keys, expected_action_keys)

    def test_info_succeed(self) -> None:
        mock_weights: ndarray[float64] = array([0.0, 0.0])
        expected_info: Dict[str, any] = {
            "action_keys": tuple(self.mock_action_keys),
            "weights": mock_weights,
        }

        info: Dict[str, any] = self.base_solver.info()

        self.assertTrue(info["action_keys"] == expected_info["action_keys"])
        self.assertTrue(
            all(
                i == j
                for i, j in zip(info["weights"], expected_info["weights"])
            )
        )



from typing import Iterable, List
import unittest
from unittest.mock import patch

from numpy import array, ndarray

from BanditAgents import actionKey

from BanditAgents.mocks.builders import BaseSolverBuilder
from BanditAgents.src.solvers.solver import Solver
from BanditAgents.src.solvers.weight_solver import WeightSolver


BASE_SOLVER_PATH: str = f"{WeightSolver.__module__}.{Solver.__name__}"
_COMPUTE_WEIGHT_PATH: str = (
    f"{WeightSolver.__module__}"
    f".{WeightSolver.__name__}"
    f".{WeightSolver._compute_weight.__name__}"
)
_RANDOM_ACTION_PATH: str = (
    f"{WeightSolver.__module__}"
    f".{WeightSolver.__name__}"
    f".{WeightSolver._random_action.__name__}"
)
_STEP_PATH: str = (
    f"{WeightSolver.__module__}"
    f".{WeightSolver.__name__}"
    f".{WeightSolver._step.__name__}"
)
_STEPS_PATH: str = (
    f"{WeightSolver.__module__}"
    f".{WeightSolver.__name__}"
    f".{WeightSolver._steps.__name__}"
)


def make_mock_base_solver(action_keys: Iterable[actionKey]) -> Solver:
    return BaseSolverBuilder().with_action_keys(action_keys).build()


def mock__steps(x: ndarray[int], y: ndarray[float]) -> Iterable[bool]:
    return (True for _ in x)


def mock__random_action() -> int:
    return 1


class TestWeightSolver(unittest.TestCase):
    mock_action_keys: List[actionKey]
    mock_optimistic_value: float
    mock_step_size: float
    weight_solver: WeightSolver

    @patch(BASE_SOLVER_PATH, side_effect=make_mock_base_solver)
    def setUp(self, mock_base_solver) -> None:
        self.mock_action_keys = ["action_a", "action_b"]
        self.mock_optimistic_value = 1.0
        self.mock_step_size = 1.0

        self.weight_solver = WeightSolver(
            action_keys=self.mock_action_keys,
            optimistic_value=self.mock_optimistic_value,
            step_size=self.mock_step_size,
        )

    def test_action_keys_to_indexes_succeed(self) -> None:
        expected_action_indexes: ndarray[int] = array([0.0, 1.0])
        action_indexes: ndarray[int] = (
            self.weight_solver.action_keys_to_indexes(self.mock_action_keys)
        )

        self.assertTrue(
            all(
                i == j for i, j in zip(action_indexes, expected_action_indexes)
            )
        )

    @patch(_STEPS_PATH, side_effect=mock__steps)
    def test_fit_succeed(self, _steps) -> None:
        self.weight_solver._steps = _steps
        mock_ids: ndarray[int] = array([0, 0, 1, 1])
        mock_targets: ndarray[float] = array([1.0, 1.0, 0.0, 0.0])

        self.weight_solver.fit(mock_ids, mock_targets)

        _steps.assert_called_with(mock_ids, mock_targets)

    @patch(_RANDOM_ACTION_PATH, side_effect=mock__random_action)
    def test_predict_succeed(self, _random_action) -> None:
        mock_weights: ndarray[float] = array([0.0, 1.0])
        self.weight_solver.weights = mock_weights
        self.weight_solver.is_trained = True

        expected_index = 1
        chosen_index: int = self.weight_solver.predict()

        self.assertEqual(chosen_index, expected_index)

    @patch(_RANDOM_ACTION_PATH, side_effect=mock__random_action)
    def test_predict_with_random_action_succeed(self, _random_action) -> None:
        self.weight_solver._random_action = _random_action

        expected_index: int = 1
        chosen_index: int = self.weight_solver.predict()

        _random_action.assert_called_once()

        self.assertEqual(chosen_index, expected_index)

    def test__init_weights_succeed(self) -> None:
        self.weight_solver._init_weights()
        initial_weights: ndarray[float] = self.weight_solver.weights
        expected_initial_weight: float = self.mock_optimistic_value

        self.assertTrue(
            all([w == expected_initial_weight for w in initial_weights])
        )

    @patch(_STEP_PATH, side_effect=lambda target, action_index: True)
    def test__steps_succeed(self, _step) -> None:
        mock_ids: ndarray[int] = array([0, 0, 1, 1])
        mock_targets: ndarray[float] = array([1.0, 1.0, 0.0, 0.0])

        self.weight_solver._step = _step
        is_success: bool = all(
            tuple(self.weight_solver._steps(mock_ids, mock_targets))
        )

        self.assertEqual(is_success, True)

    @patch(
        _COMPUTE_WEIGHT_PATH,
        side_effect=lambda weight, target, step_size_value: 1.0,
    )
    def test__step_succeed(self, _compute_weight) -> None:
        self.weight_solver._compute_weight = _compute_weight

        is_success: bool = self.weight_solver._step(1.0, 0)

        _compute_weight.assert_called_with(
            weight=self.weight_solver.weights[0],
            target=1.0,
            step_size_value=self.weight_solver.step_size,
        )

        self.assertTrue(is_success)

    def test__random_action_succeed(self) -> None:
        chosen_index = self.weight_solver._random_action()

        self.assertLessEqual(chosen_index, len(self.weight_solver.action_keys))

    def test__compute_weight_succeed(self) -> None:
        mock_weight = 1.0
        mock_target = 1.0
        mock_step_size = 1.0

        expected_weight = 1.0
        weight: float = self.weight_solver._compute_weight(
            mock_weight, mock_target, mock_step_size
        )

        self.assertEqual(weight, expected_weight)



from typing import List
import unittest
from unittest.mock import patch


from BanditAgents.src.solvers.ucb_solver import UCBSolver
from BanditAgents.src.domain import actionKey
from BanditAgents.src.solvers.weight_solver import WeightSolver

WEIGHT_SOLVER_PATH: str = f'{UCBSolver.__module__}.{WeightSolver.__name__}'
WEIGHT_SOLVER__COMPUTE_WEIGHT_PATH: str = (
    f'{WEIGHT_SOLVER_PATH}.{WeightSolver._compute_weight.__name__}'
)
_COMPUTE_WEIGHT_PATH: str = (
    f'{UCBSolver.__module__}'
    f'.{UCBSolver.__name__}'
    f'.{UCBSolver._compute_weight.__name__}'
)


def mock_weights_solver__compute_weight_succeed(
    weight: float, target: float, step_size_value: float
) -> float:
    return 1


class TestUCBSolver(unittest.TestCase):
    mock_action_keys: List[actionKey]
    mock_optimistic_value: float
    mock_step_size: float
    mock_confidence: float
    ucb_solver: UCBSolver

    @patch(WEIGHT_SOLVER_PATH)
    def setUp(self, weight_solver) -> None:
        self.mock_action_keys = ['action_a', 'action_b']
        self.mock_optimistic_value = 1, 0
        self.mock_step_size = 1.0
        self.mock_confidence = 1, 0

        self.ucb_solver = UCBSolver(
            action_keys=self.mock_action_keys,
            optimistic_value=self.mock_optimistic_value,
            step_size=self.mock_step_size,
            confidence=self.mock_confidence,
        )

    @patch(_COMPUTE_WEIGHT_PATH)
    def test__step_succeed(self, _compute_weight) -> None:
        mock_target: float = 1.0
        mock_action_index: int = 0
        self.ucb_solver._compute_weight = _compute_weight

        step_taken: bool = self.ucb_solver._step(
            mock_target, mock_action_index
        )

        self.assertEqual(_compute_weight.call_count, 5)
        self.assertTrue(step_taken)

    @patch(
        WEIGHT_SOLVER__COMPUTE_WEIGHT_PATH,
        side_effect=mock_weights_solver__compute_weight_succeed,
    )
    def test__compute_weight_succeed(self, _compute_weight) -> None:
        mock_weight = 1.0
        mock_step_size_value = 1.0
        mock_confidence = 1.0
        mock_action_count = 1
        mock_total_action_count = 1
        mock_target = 1.0

        self.ucb_solver.__class__.__bases__[0]._compute_weight = (
            _compute_weight
        )

        expected_weight: float = 1.0
        weight: float = self.ucb_solver._compute_weight(
            weight=mock_weight,
            step_size_value=mock_step_size_value,
            confidence=mock_confidence,
            action_count=mock_action_count,
            total_action_count=mock_total_action_count,
            target=mock_target,
        )

        self.assertEqual(weight, expected_weight)



from pathlib import Path

from setuptools import setup, find_packages


# read the contents of your README file

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(
    name="Bandit Agents",
    description="Library to solve k-armed bandit problems",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    setuptools_git_versioning={"enabled": True},
    setup_requires=[
        "flake8",
        'setuptools',
        'setuptools-git-versioning',
        'wheel',
    ],
)



from typing import Iterable, Self

from BanditAgents.src.domain import actionKey
from BanditAgents.src.solvers.solver import Solver


class BaseSolverBuilder:
    action_keys: Iterable[actionKey] = None

    def with_action_keys(self, action_keys: Iterable[actionKey]) -> Self:
        self.action_keys = action_keys

        return self

    def with_default_action_keys(self) -> Self:
        self.action_keys = ("action_a", "action_b")

        return self

    def build(self) -> Solver:
        if self.action_keys is None:
            self.with_default_action_keys()

        base_solver = Solver(self.action_keys)

        return base_solver



from typing import List

from BanditAgents.src.solvers.sampling_solver import SamplingSolver


class TestSamplingSolver:
    mock_action_keys: List[str]
    mock_n_sampling: int
    mock_max_sample_size: int

    sampling_solver: SamplingSolver

    def setUp(self) -> None:
        self.mock_action_keys = ['action_a', 'action_b']
        self.mock_n_sampling = 1
        self.mock_max_sample_size = 10

    def test_fit_succeed(self) -> None:
        pass

    def test_info_succeed(self) -> None:
        pass

    def test_predict_succeed(self) -> None:
        pass

    def test__sample_distribution_succeed(self) -> None:
        pass

    def test__fit_gamma_on_targets_succeed(self) -> None:
        pass



import logging
from typing import Callable, Tuple
from numpy import float64, int64, ndarray
from scipy.stats import gamma

from BanditAgents import Agent, actionKey, SyncContextHyperParameters


def basic_agent_exemple(
    actions: Tuple[Tuple[actionKey, Callable[[], float]]] = (
        ("action_a", lambda: gamma.rvs(a=6.8, scale=0.1, loc=0, size=1)[0]),
        ("action_b", lambda: gamma.rvs(a=2.2, scale=0.2, loc=0, size=1)[0]),
    ),
    n_steps: int = 100,
) -> None:
    exemple_logger: logging.Logger = logging.getLogger(__name__)

    exemple_logger.debug(
        "Starting basic agent exemple\n"
        "---------------------------------------------------\n"
        "Initiating agent"
    )

    # Now we make the agent, the default context is of type SyncContext,
    # the default solver is a SamplingSolver
    sync_context_hyperparameters = SyncContextHyperParameters(actions)
    agent: Agent = Agent(sync_context_hyperparameters)

    indexes: ndarray[int64]
    targets: ndarray[float64]

    exemple_logger.debug(f"Agent initiated \n{agent.info()}")

    for i in range(n_steps):
        exemple_logger.debug(f"running step {i}")
        indexes, targets = agent.act()

        agent = agent.fit(x=indexes, y=targets)

        exemple_logger.debug(f"agent info is: {agent.info()}")

    exemple_logger.debug("---------------------------------------------------")



from inspect import signature
from typing import Callable, Dict, Iterable, Tuple

from BanditAgents.src.contexts.deployable.deployable_context import (
    DeployableContext,
)
from BanditAgents.src.domain import actionKey


class SyncContext(DeployableContext):
    actions: Tuple[Callable[[any], float]]

    def __init__(
        self, actions: Iterable[Tuple[actionKey, Callable[[any], float]]]
    ) -> None:
        """Constructor of the SyncContext class

        The sync context is a context where the result of an action can
        be expected to be obtained instantaneously or alomost instantaneously

        Parameters
        ----------
        actions : Iterable[Tuple[actionKey, Callable[[any], float]]]
            An action consist of a pair with an actionKey at index 0
            and a callable that returns a float at index 1

            The return of the Callable will be used as
            a prediction/choice's target
        """
        self.actions: Tuple[Callable[[any], float]] = tuple(
            action for _, action in actions
        )

        super().__init__(action_key for action_key, _ in actions)

    def execute(self, action_index: int, *args, **kwargs) -> float:
        """Execute an action.

        Which meant execute actions[action_index]

        Parameters
        ----------
        action_index : int
            position of the action in the context

        Returns
        -------
        float
            target associated to the action's execution
        """
        target: float = self.actions[action_index](*args, **kwargs)

        return target

    def info(self) -> Dict[str, any]:
        context_info: Dict[str, any] = super().info()
        context_info['actions'] = [
            signature(action) for action in self.actions
        ]

        return context_info



import logging
from typing import Dict, Iterable, Self

from numpy import (
    array,
    float64,
    ndarray,
    int64,
    ones,
    vectorize,
    zeros,
    full,
    nan,
    isnan,
)
from BanditAgents.src.solvers.solver import Solver
from BanditAgents.src.domain import actionKey, solverKey
from scipy.stats import gamma


class SamplingSolver(Solver):
    action_counts: ndarray[int64]
    weights: ndarray[float64]
    max_sample_reached: ndarray[int64]
    max_sample_size: int
    n_sampling: int
    targets: ndarray[float64]

    def __init__(
        self,
        action_keys: Iterable[actionKey],
        n_sampling: int = 1,
        max_sample_size: int = 1000,
        solver_id: solverKey = None,
    ) -> None:
        """_summary_

        Parameters
        ----------
        action_keys : Iterable[actionKey]
            _description_
        n_sampling : int, optional
            _description_, by default 1
        max_sample_size : int, optional
            _description_, by default 1000
        solver_id: solverKey, optional
            _description_, by default None
        """
        self.logger: logging.Logger = logging.getLogger(__name__)

        super().__init__(action_keys=action_keys, solver_id=solver_id)

        self.action_counts = zeros(len(self.action_keys))
        self.weights = ones([3, len(self.action_keys)])
        self.max_sample_size = max_sample_size
        self.max_sample_reached = zeros(len(self.action_keys))
        self.n_sampling = n_sampling
        self.targets = full((self.max_sample_size, len(self.action_keys)), nan)

    def fit(self, x: ndarray[int], y: ndarray[float]) -> Self:
        """_summary_

        Parameters
        ----------
        x : ndarray[int]
            _description_
        y : ndarray[float]
            _description_

        Returns
        -------
        Self
            _description_
        """
        self.logger.debug("fitting sample solver")
        assert x.size == y.size

        for action_index, target in zip(x, y):
            self.targets[
                int(self.action_counts[action_index]), action_index
            ] = target

            if self.action_counts[action_index] < self.max_sample_size - 1:
                self.action_counts[action_index] += 1
            else:
                self.action_counts[action_index] = 0
                self.max_sample_reached[action_index] += 1

        self.logger.debug(self.targets.shape)

        for i in range(self.targets.shape[1]):
            self.weights[:, i] = self._fit_gamma_on_targets(self.targets[:, i])

        return self

    def info(self) -> Dict[str, any]:
        info = super().info()
        info["targets"] = self.targets

        return info

    def predict(self) -> int:
        """_summary_

        Returns
        -------
        int
            _description_
        """

        def sample_distribution(
            alpha: float, loc: float, scale: float
        ) -> float:
            mean_sample: float = self._sample_distribution(
                alpha=alpha, loc=loc, scale=scale
            )

            return mean_sample

        vec_sample_distribution = vectorize(sample_distribution)

        samples: ndarray[float64] = vec_sample_distribution(
            self.weights[0, :], self.weights[1, :], self.weights[2, :]
        )

        return samples.argmax()

    def _sample_distribution(self, alpha, loc, scale) -> float:
        """_summary_

        Parameters
        ----------
        alpha : _type_
            _description_
        loc : _type_
            _description_
        scale : _type_
            _description_

        Returns
        -------
        float
            _description_
        """
        samples: ndarray[float64] = gamma.rvs(
            a=alpha, loc=loc, scale=scale, size=self.n_sampling
        )

        mean_sample: float = sum(samples) / self.n_sampling

        return mean_sample

    def _fit_gamma_on_targets(
        self, targets: ndarray[float64]
    ) -> ndarray[float64]:
        """_summary_

        Parameters
        ----------
        targets : ndarray[float64]
            _description_

        Returns
        -------
        ndarray[float64]
            _description_
        """
        training_targets = targets[~isnan(targets)]

        if training_targets.var() > 0:
            shape, loc, scale = gamma.fit(training_targets)

        else:
            shape, loc, scale = 1, 1, 1

        return array([shape, loc, scale])



from typing import Iterable

from numpy import (
    float64,
    full,
    int64,
    ndarray,
    vectorize,
    zeros,
    sqrt,
    log,
    inf,
)
from BanditAgents.src.solvers.weight_solver import WeightSolver
from BanditAgents.src.domain import actionKey, solverKey


class UCBSolver(WeightSolver):
    action_counts: ndarray[int64]
    confidence: float64
    total_action_count: int

    def __init__(
        self,
        action_keys: Iterable[actionKey],
        optimistic_value: float = 0.0,
        step_size: float = 1.0,
        confidence: float = 1.0,
        solver_id: solverKey = None,
    ) -> None:
        """_summary_

        Parameters
        ----------
        action_keys : Iterable[actionKey]
            _description_
        optimistic_value : _type_, optional
            _description_, by default 0.
        step_size : _type_, optional
            _description_, by default 1.
        confidence : float, optional
            _description_, by default 1.
        solver_id: solverKey, optional
            _description_, by default None
        """
        super().__init__(
            action_keys=action_keys,
            optimistic_value=optimistic_value,
            step_size=step_size,
            solver_id=solver_id,
        )

        self.action_counts = zeros(len(self.action_keys))
        self.confidence = confidence
        self.total_action_count = 0

    def _step(self, target: float, action_index: int) -> bool:
        """_summary_

        Parameters
        ----------
        target : float
            _description_
        action_index : int
            _description_

        Returns
        -------
        bool
            _description_
        """
        targets: ndarray[float64] = full(len(self.action_keys), -inf)
        targets[action_index] = target

        self.action_counts[action_index] += 1
        self.total_action_count += 1

        def compute_weight(
            action_weight: float, action_target: float, action_count: int
        ) -> float:
            new_weight: float = self._compute_weight(
                weight=action_weight,
                step_size_value=self.step_size,
                confidence=self.confidence,
                action_count=action_count,
                total_action_count=self.total_action_count,
                target=action_target,
            )

            return new_weight

        vec_compute_weight = vectorize(compute_weight)

        self.weights = vec_compute_weight(
            self.weights, targets, self.action_counts
        )

        return True

    @staticmethod
    def _compute_weight(
        weight: float,
        step_size_value: float,
        confidence: float,
        action_count: int,
        total_action_count: int,
        target: float = None,
    ) -> float:
        """_summary_

        Parameters
        ----------
        weight : float
            _description_
        step_size_value : float
            _description_
        confidence : float
            _description_
        action_count : int
            _description_
        total_action_count : int
            _description_
        target : float, optional
            _description_, by default None

        Returns
        -------
        float
            _description_
        """
        if target > -inf and target < inf:
            new_weight: float = WeightSolver._compute_weight(
                weight=weight, target=target, step_size_value=step_size_value
            )

        else:
            new_weight: float = weight

        upper_bound_boost_numerator: float = log(total_action_count)
        upper_bound_boost_denominator: float = (
            action_count if action_count > 0 else 1.0
        )
        upper_bound_boost: float = sqrt(
            upper_bound_boost_numerator / upper_bound_boost_denominator
        )
        new_weight = new_weight + (confidence * upper_bound_boost)

        return new_weight
